@@ -9,7 +9,7 @@
 #' @aliases dkdecop pkdecop rkdecop
 #'
 #' @param u \code{mx2} matrix of evaluation points.
-#' @param obj \code{kdecop} object.
+#' @param obj \code{kdecopula} object.
 #' @param stable logical; option for stabilizing the estimator: the estimated
 #' density is cut off at \eqn{50}.
 #' 
@@ -60,25 +60,25 @@
 #'
 dkdecop <- function(u, obj, stable = FALSE) {
     stopifnot(is.numeric(u))
-    stopifnot(all(u > 0 & u < 1))
+    stopifnot(all(u >= 0 & u <= 1))
     stopifnot(inherits(obj, "kdecopula"))
     stopifnot(is.logical(stable))
-
+    
     ## define appropriately shaped udata frame for vapply
     u <- as.matrix(u)
     if (ncol(u) == 1)
         u <- matrix(u, 1L, nrow(u))
-
+    
     ## adjust for flipping option of kdevine package
     d <- ncol(u)
     if (!is.null(obj$flip))
         u <- matrix(u[, 2:1], nrow(u), d)
-
-
+    
+    
     ## if independence copula is specified return 1
     if ("indep.copula" %in% class(obj))
         return(rep(1, nrow(u)))
-
+    
     ## evaluate density  (use faster algorithm for d = 2)
     u <- pmin(pmax(u, 1e-3), 1 - 1e-3)
     if (d == 2) {
@@ -89,17 +89,17 @@ dkdecop <- function(u, obj, stable = FALSE) {
         # define help indicators
         tmplst <- split(rep(seq(-1, 2, 1), d), ceiling(seq.int(4*d)/4))
         helpind <- as.matrix(do.call(expand.grid, tmplst))
-
+        
         out <- interp(u,
                       obj$estimate,
                       obj$grid,
                       helpind)
     }
-
+    
     ## stabilize output
     if (stable)
         out <- pmin(out, 10^(1 + d/2))
-
+    
     ## return results
     out
 }
@@ -107,7 +107,7 @@ dkdecop <- function(u, obj, stable = FALSE) {
 #' @rdname dkdecop
 #'
 pkdecop <- function(u, obj) {
-    stopifnot(all(u > 0 & u < 1))
+    stopifnot(all(u >= 0 & u <= 1))
     stopifnot(inherits(obj, "kdecopula"))
     ## define appropriately shaped u matrix
     u <- as.matrix(u)
@@ -119,18 +119,18 @@ pkdecop <- function(u, obj) {
         cond.var <- ifelse(cond.var == 1, 2, 1)
     }
     d <- ncol(u)
-
+    
     ## if independence copula is specified, return prod(u) directly
     if ("indep.copula" %in% class(obj))
         return(apply(u, 1, prod))
-
+    
     ## define help objects
     tmplst <- split(rep(seq(-1, 2, 1), d), ceiling(seq.int(4*d)/4))
     helpind <- as.matrix(do.call(expand.grid, tmplst))
     m <- length(obj$grid)
     tmplst <- split(rep(obj$grid, d), ceiling(seq.int(m*d)/m))
     helpgrid <- as.matrix(do.call(expand.grid, tmplst))
-
+    
     ## evaluate cdf
     eval_cdf(u,
              obj$estimate,
@@ -150,7 +150,7 @@ rkdecop <- function(n, obj, quasi = FALSE) {
     n <- round(n)
     stopifnot(inherits(obj, "kdecopula"))
     stopifnot(is.logical(quasi))
-
+    
     if (!quasi) {
         # simulate independent uniform random variables
         W <- cbind(runif(n), runif(n))
@@ -158,75 +158,113 @@ rkdecop <- function(n, obj, quasi = FALSE) {
         # generate quasi random numbers
         W <- ghalton(n, d = 2)
     }
-
+    
     # if independence copula is specified, return W
     if("indep.copula" %in% class(obj))
         return(W)
-
+    
     # invert h-function otherwise
     U2 <- inv_hfunc(W,
                     1L,
                     obj$estimate,
                     obj$grid)
-
+    
     ## return results
     out <- cbind(W[, 1], U2)
     colnames(out) <- NULL
     out
 }
 
-# Evaluate h-function of kdecopula object (internal for kdevine-package)
-#
-# @param cond.var integer; either 1 or 2; specifies on which argument to
-#  condition
-#
-hkdecop <- function(u, obj, cond.var) {
-    stopifnot(all(u > 0) & all(u < 1))
+#' H-function and inverse of kdecopula object
+#' 
+#' Evaluates the h-function (or its inverse) corresponding to a \code{kdecopula}
+#' object. H-functions are conditional distribution functions obtained by
+#' integrating the copula density w.r.t. to one of its arguments (see also
+#' \code{\link[VineCopula:BiCopHfunc]{BiCopHfunc}}).
+#' 
+#' @param u \eqn{n x 2} matrix of evaluation points.
+#' @param obj \code{kdecopula} object.
+#' @param cond.var integer; \code{cond.var = 1} conditions on the first variable,
+#' \code{cond.var = 2} on the secon.
+#' @param inverse logical; indicates wether the h-function or its inverse shall be
+#' calculated.
+#' 
+#' @return A length \eqn{n} vector of the (inverse) h-function evaluated at
+#' \code{u}.
+#' 
+#' @author Thomas Nagler
+#' 
+#' @seealso 
+#' \code{\link[kdecopula:kdecop]{kdecop}},
+#' \code{\link[VineCopula:BiCopHfunc]{BiCopHfunc}}
+#' 
+#' @examples 
+#' ## load data and transform with empirical cdf
+#' data(wdbc)
+#' udat <- apply(wdbc[, -1], 2, function(x) rank(x)/(length(x)+1))
+#' 
+#' ## estimation of copula density of variables 5 and 6
+#' dens.est <- kdecop(udat[, 5:6])
+#' plot(dens.est) 
+#' 
+#' ## evaluate h-function estimate and its inverse at (u1|u2) = (0.123 | 0.321)
+#' hkdecop(c(0.123, 0.321), dens.est, cond.var = 2) 
+#' hkdecop(c(0.123, 0.321), dens.est, cond.var = 2, inverse = TRUE) 
+#' 
+hkdecop <- function(u, obj, cond.var, inverse = FALSE) {
+    stopifnot(all(u >= 0) & all(u <= 1))
     stopifnot(inherits(obj, "kdecopula"))
     stopifnot(all(cond.var %in% c(1, 2)))
     ## define appropriately shaped u matrix
     u <- as.matrix(u)
     if(ncol(u) == 1)
         u <- matrix(u, 1L, nrow(u))
-    # adjust for flipping option of kdevine package
+    ## adjust for flipping option of kdevine package
     if(!is.null(obj$flip)) {
         u <- matrix(u[, 2:1], nrow(u))
         cond.var <- ifelse(cond.var == 1, 2, 1)
     }
     d <- ncol(u)
-
-    ## if independence copula is specified, return the conditioned variable
+    
+    # if independence copula is specified, return the conditioned variable
     if ("indep.copula" %in% class(obj))
         return(u[, -cond.var])
-
-    ## evaluate h-function otherwise
-    if (d == 2) {
-        # faster algorithm for d = 2
-        return(eval_hfunc_2d(u,
+    
+    if (!inverse) {
+        # h-function
+        out <- eval_hfunc_2d(u,
                              as.integer(cond.var),
                              obj$estimate,
-                             obj$grid))
+                             obj$grid)
+        # # define help objects
+        # tmplst <- split(rep(seq(-1, 2, 1), d), ceiling(seq.int(4*d)/4))
+        # helpind <- as.matrix(do.call(expand.grid, tmplst))
+        # tmplst <- lapply(1:d,
+        #                  prep_hfunc,
+        #                  cond.var = cond.var,
+        #                  grid = obj$grid,
+        #                  d = d)
+        # helpgrid <- as.matrix(do.call(expand.grid, tmplst))
+        # uncond.var <- seq.int(d)[-cond.var]
+        # # call routine
+        # out <- eval_hfunc(u,
+        #                   as.integer(cond.var),
+        #                   as.integer(uncond.var),
+        #                   obj$estimate,
+        #                   obj$grid,
+        #                   helpgrid,
+        #                   helpind)
     } else {
-        # define help objects
-        tmplst <- split(rep(seq(-1, 2, 1), d), ceiling(seq.int(4*d)/4))
-        helpind <- as.matrix(do.call(expand.grid, tmplst))
-        tmplst <- lapply(1:d,
-                         prep_hfunc,
-                         cond.var = cond.var,
-                         grid = obj$grid,
-                         d = d)
-        helpgrid <- as.matrix(do.call(expand.grid, tmplst))
-        uncond.var <- seq.int(d)[-cond.var]
-
-        # call routine
-        return(eval_hfunc(u,
-                          as.integer(cond.var),
-                          as.integer(uncond.var),
-                          obj$estimate,
-                          obj$grid,
-                          helpgrid,
-                          helpind))
+        # inverse h-function
+        out <- inv_hfunc(u,
+                         cond.var,
+                         obj$estimate,
+                         obj$grid)
     }
+    
+    ## return results
+    colnames(out) <- NULL
+    out
 }
 
 ## Prepare the list for constructing the helpgrid in hkdecop
@@ -237,28 +275,3 @@ prep_hfunc <- function(i, cond.var, grid, d) {
         return(grid)
     }
 }
-
-## Calculate inverse of h-function
-hinvkdecop <- function(u, obj, cond.var) {
-    stopifnot(all(u > 0 & u < 1))
-    stopifnot(inherits(obj, "kdecopula"))
-    stopifnot(all(cond.var %in% c(1, 2)))
-    ## simulate independent uniform random variables
-    W <- u
-
-    # if independence copula is specified, return W
-    if("indep.copula" %in% class(obj))
-        return(W)
-
-    ## invert h-function otherwise
-    U2 <- inv_hfunc(W,
-                    cond.var,
-                    obj$estimate,
-                    obj$grid)
-
-    ## return results
-    out <- cbind(W[, 1], U2)
-    colnames(out) <- NULL
-    out
-}
-
